/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/PartitionOverride/RCS/PartitionOverride.cpp,v 1.4 2007/02/04 22:01:58 agmsmith Exp $
 *
 * This is a command line program for getting and setting the partition
 * size and offset of a device file that represents a partition on a disk,
 * thus letting you mount that partition without relying on the standard
 * partition table loading system.
 *
 * I wrote this to help with my new 320GB hard drive.  BeOS R5 only recognises
 * up to 128GB of space.  So I partitioned the drive with the BFS partition
 * taking up the the first 128GB, which worked well enough until I added a
 * efs2 partition for the rest of the space for Linux.  The BeOS partition
 * loader balked at that - claming the partititon table had a partition going
 * past the end of the disk, and thus didn't load any of the partitions.  So
 * now I'm writing this little program to replace the BeOS partition loader's
 * functionality manually.
 *
 * Partitioning tips: To wipe the old partition table out and start fresh with
 * the most compatible partition table possible (if you don't want to use the
 * existing partitioning), clear the beginning of the disk to zero with a
 *
 * cat /dev/zero >/dev/disk/ide/ata/2/slave/0/raw
 *
 * command (change the /dev part to the disk you're working on), stop it after
 * a few seconds with control-C.  Boot up in DOS and use fdisk (which will
 * report some odd disk size) to make a new small DOS partition, 100MB is good
 * enough.  This creates a standard partition table that everyone can read.
 * Then boot BeOS and make the BFS partition taking up the rest of the space.
 * Then mount the DOS partition and the new BFS one and use this program to get
 * their starting offset and sizes.  Then boot up Linux and make the partition
 * for the rest of the space (starting after the 128GB boundary to be extra
 * safe) and format it with efs2 or whatever.  Then boot up BeOS (your new
 * partition's won't show up) and add this program to the startup sequence to
 * set up the partitions and add a few mount commands to make the new BFS and
 * DOS volumes visible.  For example, my /boot/home/config/boot/UserBootscript
 * file has these lines in it:
 *
 * #PartitionManualOverride /dev/disk/ide/ata/2/master/0/0_0  106896384 32256 0 "/dev/disk/ide/ata/2/master/0/raw" 512 0
 * #mkdir /dos320gb
 * #mount -t dos /dev/disk/ide/ata/2/master/0/0_0 /dos320gb
 *
 * PartitionManualOverride /dev/disk/ide/ata/2/master/0/0_1 137329274880 106928640 1 "/dev/disk/ide/ata/2/master/0/raw" 512 0
 * mkdir /CommonBe
 * mount -t bfs /dev/disk/ide/ata/2/master/0/0_1 /CommonBe
 *
 * $Log: PartitionOverride.cpp,v $
 * Revision 1.4  2007/02/04 22:01:58  agmsmith
 * Spelling corrections and documentation improvements.
 *
 * Revision 1.3  2007/02/04 21:25:01  agmsmith
 * First working version.
 *
 * Revision 1.2  2007/01/27 17:58:36  agmsmith
 * Under construction.
 *
 * Revision 1.1  2007/01/27 16:45:37  agmsmith
 * Initial revision
 */

/* Posix headers. */

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <SupportDefs.h>
#include <Drivers.h>
#include <StorageKit.h>

#define PROGRAM_NAME_WORD "PartitionManualOverride"


/******************************************************************************
 * Global utility function to print an error message to stderr and return.  The
 * message part describes the error, and if ErrorNumber is non-zero, gets the
 * string ", error code $X (standard description)." appended to it.  If the
 * message is NULL then it gets defaulted to "Something went wrong".  The title
 * part defaults to "Error Message" if you didn't specify one.
 */

static void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  char ErrorBuffer [PATH_MAX + 1500];

  if (TitleString == NULL)
    TitleString = PROGRAM_NAME_WORD " Error Message";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occured.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

  fprintf (stderr, "%s: %s\n", TitleString, MessageString);
}



/******************************************************************************
 * Print a partition's settings to standard output.
 */
 
void PrintPartition (partition_info *PartitionInfo)
{
  printf ("Size, Offset, Partition, RawPath, BlockSize, Session.\n"
    "%qd %qd %d \"%s\" %d %d\n",
    PartitionInfo->size,
    PartitionInfo->offset,
    (int) PartitionInfo->partition,
    PartitionInfo->device,
    (int) PartitionInfo->logical_block_size,
    (int) PartitionInfo->session);
}



/******************************************************************************
 * Open the given file and use the ioctl to get partition settings from it.
 * Returns true if successful.
 */

static bool ShowPartition (const char *PartitionFileName)
{
  int             DeviceFileDescriptor;
  partition_info  PartitionInfo;
  bool            ReturnCode = false;

  printf ("Showing settings for partition file \"%s\":\n", PartitionFileName);

  DeviceFileDescriptor = open (PartitionFileName, O_RDONLY,
      0777 /* access mode flags, rwxrwxrwx is 777 in octal. */);
  if (DeviceFileDescriptor < 0)
  {
    DisplayErrorMessage ("Unable to open partition file", errno);
    return false;
  }
  
  memset (&PartitionInfo, 0, sizeof (PartitionInfo));

  if (ioctl (DeviceFileDescriptor, B_GET_PARTITION_INFO,
    &PartitionInfo, sizeof (PartitionInfo)) != 0)
  {
    DisplayErrorMessage (
      "IO Control operation B_GET_PARTITION_INFO has failed", errno);
  }
  else
  {
    ReturnCode = true;
    PrintPartition (&PartitionInfo);
  }

  close (DeviceFileDescriptor);
  return ReturnCode;
}



/******************************************************************************
 * Create and set up a partition file.
 */

static bool SetPartition (
  const char *PartitionFileName,
  partition_info *PartitionInfo)
{
  int DeviceFileDescriptor;
  bool ReturnCode = false;

  printf ("Creating a partition file named \"%s\":\n", PartitionFileName);
  PrintPartition (PartitionInfo);

  DeviceFileDescriptor = open (PartitionFileName, O_CREAT | O_RDONLY,
      0777 /* access mode flags, rwxrwxrwx is 777 in octal. */);
  if (DeviceFileDescriptor < 0)
  {
    DisplayErrorMessage ("Unable to create a partition file", errno);
    return false;
  }

  if (ioctl (DeviceFileDescriptor, B_SET_PARTITION,
    PartitionInfo, sizeof (*PartitionInfo)) != 0)
  {
    DisplayErrorMessage (
      "IO Control operation B_SET_PARTITION has failed", errno);
  }
  else
  {
    ReturnCode = true;
    printf ("Done.\n");
  }

  close (DeviceFileDescriptor);
  return ReturnCode;
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char** argv)
{
  partition_info PartitionInfo;
  char *PartitionPath;
  char *StringPntr;

  if (argc <= 1) {
    printf ("Usage: %s PartitionPath [Size [Offset [Partition [RawPath [BlockSize [Session]]]]]]\n\n", argv[0]);
    printf (
"A program to manipulate partition files in BeOS, by Alexander G. M. Smith.\n"
"\n"
"PartitionPath is something like /dev/disk/ide/ata/2/slave/0/0_3 where 0_3 means\n"
"session 0 (only non-zero for CD-ROMs) and partition 3 (zero based numbering).\n"
"\n"
"If you just specify the PartitionPath, it will read that existing partition\n"
"file and print out its settings in a form you can copy and paste to make a\n"
"command to recreate that partition file.\n"
"\n"
"If you also specify the size (in decimal bytes), it will create the\n"
"corresponding partition file.  If not specified, the offset is zero, block\n"
"size will default to 512, the session, partition and raw device will default\n"
"to whatever it can dig out from the partition path or zero if it can't figure\n"
"it out.\n"
"\n"
"See source code for partitioning tips.  Please be careful, you can easily\n"
"wipe out your data accidentally if you don't do everything perfectly!\n"
"\n"
PROGRAM_NAME_WORD " $Revision: 1.4 $ compiled on "
      __DATE__ " at " __TIME__ ".\n\n");
    return 0;
  }

  // Set up some default values, replace with user's arguments if available.

  memset (&PartitionInfo, 0, sizeof (PartitionInfo));

  PartitionPath = argv[1];

  PartitionInfo.logical_block_size = 512;

  // Make up the raw device name from the parition one by replacing the path
  // ending with "/raw".

  strncpy (PartitionInfo.device, PartitionPath,
    sizeof (PartitionInfo.device) - 5 /* space to append "/raw" */ );
  StringPntr = strrchr (PartitionInfo.device, '/');
  if (StringPntr == NULL)
    strcat (PartitionInfo.device, "/raw");
  else
    strcpy (StringPntr, "/raw");

  // Get the session and partition numbers from the file name portion of the
  // partition path.  The thing that looks like 1_2 means session 1,
  // partition 2.

  StringPntr = strrchr (PartitionPath, '/');
  if (StringPntr == NULL)
    StringPntr = PartitionPath;
  PartitionInfo.session = atol (StringPntr);
  StringPntr = strchr (StringPntr, '_');
  if (StringPntr != NULL)
    PartitionInfo.partition = atol (StringPntr);

  // Now that defaults have been set, overwrite them with user provided values
  // where they exist.

  if (argc >= 3)
    PartitionInfo.size = atoll (argv[2]);

  if (argc >= 4)
    PartitionInfo.offset = atoll (argv[3]);

  if (argc >= 5)
    PartitionInfo.partition = atol (argv[4]);

  if (argc >= 6) {
    strncpy (PartitionInfo.device, argv[5], sizeof (PartitionInfo.device));
    PartitionInfo.device[sizeof(PartitionInfo.device)-1] = 0;
  }

  if (argc >= 7)
    PartitionInfo.logical_block_size = atol (argv[6]);

  if (argc >= 8)
    PartitionInfo.session = atol (argv[7]);

  if (PartitionInfo.size <= 0 || PartitionInfo.offset < 0)
    return ShowPartition (PartitionPath);
  else
    return SetPartition (argv[1], &PartitionInfo);

  return 0;
}
